/******************************************************************************
 * The MIT License (MIT)
 *
 * Copyright (c) 2019-2026 Baldur Karlsson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 ******************************************************************************/

#pragma once

#include "api/replay/rdcstr.h"

namespace StringFormat
{
struct Args
{
  virtual void reset() = 0;

  virtual void error(const char *err) = 0;

  virtual int get_int() = 0;
  virtual unsigned int get_uint() = 0;
  virtual double get_double() = 0;
  virtual void *get_ptr() = 0;
  virtual const char *get_str() = 0;
  virtual uint64_t get_uint64() = 0;
  size_t get_size() { return sizeof(size_t) == 8 ? (size_t)get_uint64() : get_uint(); }
};

rdcstr Fmt(const char *format, ...);
rdcstr Fmt(rdcliteral format, ...);
rdcstr FmtArgs(const char *format, Args &args);
};
